! T3XFORTH Bootstrapping Compiler
! Nils M Holm, 2021
! Public domain / CC0 License

module tfcmp(t3x);

object	t[t3x];

const	MEMTOP	= 16384;
const	BUFLEN	= 1024;
const	STKLEN	= 256;
const	ORG	= 0x100;
const	IDLEN	= 5;

const	COMP	= 0x40;
const	IMMED	= 0x80;

const	NAME	= 0,
	LFA	= 6,
	CFA	= 8,
	PFA	= 10;

var	Mem::MEMTOP;
var	Here;
var	Coldjmp;

var	Buffer::BUFLEN;
var	Nchars, Bp;

var	Stack[STKLEN];
var	Sp;

var	Last;
var	DoLit;
var	DoColon;
var	DoVar;
var	DoZBranch;
var	DoBranch;
var	Exit;
var	DoDo;
var	DoLoop;
var	DoPlusLoop;
var	DoUpLoop;
var	DoString;
var	DoAbortS;
var	DoConst;
var	Dict;

var	Input;

str_length(s) return t.memscan(s, 0, 32767);

str_equal(s1, s2)
	return t.memcomp(s1, s2, str_length(s1)+1) = 0;

writes(s) t.write(T3X.SYSOUT, s, str_length(s));

nl() do var b::3;
	writes(t.newline(b));
end

var ntoa_buf::100;

ntoa(x) do var i, k;
	if (x = 0) return "0";
	i := 0;
	k := x<0-> -x: x;
	while (k > 0) do
		i := i+1;
		k := k/10;
	end
	i := i+1;
	if (x < 0) i := i+1;
	ntoa_buf::i := 0;
	k := x<0-> -x: x;
	while (k > 0) do
		i := i-1;
		ntoa_buf::i := '0' + k mod 10;
		k := k/10;
	end
	if (x < 0) do
		i := i-1;
		ntoa_buf::i := '-';
	end
	return @ntoa_buf::i;
end

error(m, s) do var b::3;
	t.write(T3X.SYSERR, "error: ", 7);
	t.write(T3X.SYSERR, m, str_length(m));
	if (s) do
		t.write(T3X.SYSERR, ": ", 2);
		t.write(T3X.SYSERR, s, str_length(s));
	end
	t.newline(b);
	t.write(T3X.SYSERR, b, str_length(b));
	halt 1;
end

mword(a) return Mem::(a+1) << 8 | Mem::a;

loadimage(src) do var fd;
	fd := t.open(src, T3X.OREAD);
	if (fd < 0) error("image not found", src);
	Here := t.read(fd, Mem, MEMTOP);
	t.close(fd);
	Here := Here - 2;
	Dict := mword(Here);
	Coldjmp := Here-5;
end

saveimage(dest) do var fd, k;
	fd := t.open(dest, T3X.OWRITE);
	if (fd < 0) error("cannot create image", 0);
	k := t.write(fd, Mem, Here);
	t.close(fd);
	if (k \= Here) error("file write error", 0);
end

listwords() do var p, n;
	p := Dict-ORG;
	n := 0;
	while (p \= -ORG) do
		t.write(T3X.SYSOUT, @Mem::(p+1), IDLEN);
		writes("\s\s\s\s\s\s\s\s"+IDLEN);
		n := n+1;
		if (n /\ n mod 8 = 0) nl();
		p := mword(p+LFA)-ORG;
	end
	nl();
	writes(ntoa(n));
	writes(" WORDS, ");
	writes(ntoa(Here));
	writes(" BYTES");
	nl();
end

rdchar() do
	if (Bp >= Nchars) do
		Nchars := t.read(Input, Buffer, BUFLEN);
		if (Nchars < 1) return %1;
		Bp := 0;
	end
	Bp := Bp+1;
	return Buffer::(Bp-1);
end

space(c) return c = '\s' \/ c = '\n' \/ c = '\r';

var	Wordbuf::64;

word(d) do var i, c;
	c := rdchar();
	while (space(c)) c := rdchar();
	if (c = %1) return %1;
	i := 0;
	while (1) do
		if (d = '\s' /\ space(c)) leave;
		if (d = c \/ d = %1) leave;
		if (i < 64) do
			Wordbuf::i := c;
			i := i+1;
		end
		c := rdchar();
	end
	Wordbuf::i := 0;
	return Wordbuf;
end

emitb(b) do
	Mem::Here := b;
	Here := Here+1;
end

emitw(w) do
	emitb(w & 255);
	emitb(w >> 8 & 255);
end

patch(a, w) do
	Mem::a := w & 255;
	Mem::(a+1) := w >> 8 & 255;
end

create(w) do var k, i;
	Last := Here;
	emitb(str_length(w) & 0x1F);
	k := str_length(w);
	for (i=0, IDLEN) emitb(i<k-> w::i: '\s');
	emitw(Dict);
	emitw(DoVar);
end

digit(c) return '0' <= c /\ c <= '9';

number(w, pv) do var s, i, v;
	s := 1;
	i := 0;
	v := 0;
	if (w::i = '-') do
		i := 1;
		s := %1;
	end
	while (w::i) do
		if (\digit(w::i)) return 0;
		v := v * 10 + w::i - '0';
		i := i+1;
	end
	pv[0] := v*s;
	return %1;
end

tocode(w) return w - PFA + CFA;

find(w, pv) do var p, i, k, b::IDLEN+1;
	p := Dict - ORG;
	k := str_length(w);
	t.memcopy(b, w, IDLEN);
	for (i=k, IDLEN) b::i := '\s';
	b::IDLEN := 0;
	while (p \= -ORG) do
		if (	Mem::p & 0x1F = k /\
			t.memcomp(b, @Mem::(p+1), IDLEN) = 0
		) do
			if (pv) pv[0] := p + PFA + ORG;
			return p + PFA + ORG;
		end
		p := mword(p+LFA)-ORG;
	end
end

push(w) do
	if (Sp >= STKLEN) error("stack overflow", 0);
	Stack[Sp] := w;
	Sp := Sp+1;
end

pop() do
	if (Sp < 1) error("stack underflow", 0);
	Sp := Sp-1;
	return Stack[Sp];
end

mark() do
	emitw(DoZBranch);
	push(Here);
	emitw(0);
end

resolve() patch(pop(), here+ORG);

compelse() do var a;
	emitw(DoBranch);
	a := Here;
	emitw(0);
	resolve();
	push(a);
end

emitlit(x) do
	emitw(DoLit);
	emitw(x);
end

charlit() do var c;
	c := word('\s');
	emitlit(c::0);
end

compdo() do
	emitw(DoDo);
	push(Here);
end

comprepeat() do var a;
	a := pop();
	emitw(DoBranch);
	emitw(pop()+ORG);
	patch(a, here+ORG);
end

back(op) do
	emitw(op);
	emitw(pop()+ORG);
end

notfound(w) error("not found", w);

compimmed() do var a, w;
	w := word('\s');
	if (find(w, @a) = 0) notfound(w);
	emitw(tocode(a));
end

lookup(w, code) do var a;
	a := find(w, 0);
	if (a = 0) notfound(w);
	return code-> tocode(a): a;
end

compstring() do var w, k, i;
	if (DoString = 0) DoString := lookup("($\q)", 1);
	w := word('"');
	k := str_length(w);
	emitw(DoString);
	emitb(k);
	for (i=0, k) emitb(w::i);
end

compabort() do var w, k, i;
	if (DoAbortS = 0) DoAbortS := lookup("(ABORT\q)", 1);
	w := word('"');
	k := str_length(w);
	emitw(DoAbortS);
	emitb(k);
	for (i=0, k) emitb(w::i);
end

hexdigit(c) return 'A' <= c /\ c <= 'F';

hex(w) do var i, v;
	v := 0;
	i := 0;
	while (w::i) do
		     ie (digit(w::i)) v := v*16 + w::i-'0';
		else ie (hexdigit(w::i)) v := v*16 + w::i-'A'+10;
		else error("bad hex number", w);
		i := i+1;
	end
	return v;
end

comment() do var c;
	c := rdchar();
	while (c \= '\n' /\ c \= %1) c := rdchar();
end

compaddr() do var w, a;
	w := word('\s');
	a := find(w, 0);
	if (a = 0) notfound(w);
	emitlit(a);
end

emithex() emitlit(hex(word('\s')));

compile() do var w, x;
	w := word('\s');
	create(w);
	Here := Here-2;
	emitw(DoColon);
	w := word('\s');
	while (\str_equal(w, ";")) do
		     ie (str_equal(w, "(")) word(')');
		else ie (str_equal(w, "$\q")) compstring();
		else ie (str_equal(w, "ABORT\q")) compabort();
		else ie (str_equal(w, "AGAIN")) back(DoBranch);
		else ie (str_equal(w, "BEGIN")) push(Here);
		else ie (str_equal(w, "DO")) compdo();
		else ie (str_equal(w, "ELSE")) compelse();
		else ie (str_equal(w, "IF")) mark();
		else ie (str_equal(w, "LOOP")) back(DoLoop);
		else ie (str_equal(w, "+LOOP")) back(DoPlusLoop);
		else ie (str_equal(w, "/LOOP")) back(DoUpLoop);
		else ie (str_equal(w, "REPEAT")) comprepeat();
		else ie (str_equal(w, "THEN")) resolve();
		else ie (str_equal(w, "UNTIL")) back(DoZBranch);
		else ie (str_equal(w, "WHILE")) mark();
		else ie (str_equal(w, "[']")) compaddr();
		else ie (str_equal(w, "[CHAR]")) charlit();
		else ie (str_equal(w, "[COMPILE]")) compimmed();
		else ie (str_equal(w, "[HEX]")) emithex();
		else ie (str_equal(w, "\\")) comment();
		else ie (find(w, @x)) emitw(tocode(x));
		else ie (number(w, @x)) emitlit(x);
		else notfound(w);
		w := word('\s');
	end
	Dict := Last + ORG;
	emitw(Exit);
end

pushaddr() do var w, a;
	w := word('\s');
	a := find(w, 0);
	if (a = 0) notfound(w);
	push(a);
end

setflag(f) Mem::Last := Mem::Last | f;

adjust(n) push(pop()+n);

store() do var a;
	a := pop() - ORG;
	patch(a, pop());
end

defer() do
	create(word('\s'));
	Here := Here-2;
	emitw(DoColon);
	emitw(0);
	emitw(Exit);
	Dict := Last + ORG;
end

undefer() do var a, w;
	w := word('\s');
	if (find(w, @a) = 0) notfound(w);
	patch(a-ORG, Last+ORG+CFA);
end

variable() do
	create(word('\s'));
	emitw(0);
	Dict := Last + ORG;
end

constant() do
	create(word('\s'));
	Here := Here-2;
	emitw(DoConst);
	emitw(pop());
	Dict := Last + ORG;
end

interpret(w) do var x;
	     ie (str_equal(w, "(")) word(')');
	else ie (str_equal(w, ":")) compile();
	else ie (str_equal(w, "'")) pushaddr();
	else ie (str_equal(w, "!")) store();
	else ie (str_equal(w, ",")) emitw(pop());
	else ie (str_equal(w, "COMPILE-ONLY")) setflag(COMP);
	else ie (str_equal(w, "CONSTANT")) constant();
	else ie (str_equal(w, "DEFER")) defer();
	else ie (str_equal(w, "IMMEDIATE")) setflag(IMMED);
	else ie (str_equal(w, "IS")) undefer();
	else ie (str_equal(w, ">NFA")) push(pop()-PFA);
	else ie (str_equal(w, "VARIABLE")) variable();
	else ie (str_equal(w, "[HEX]")) push(hex(word('\s')));
	else ie (str_equal(w, "\\")) comment();
	else ie (number(w, @x)) push(x);
	else error("cannot interpret", w);
end

loadsrc() do var w;
	w := word('\s');
	while (w \= %1) do
		interpret(w);
		w := word('\s');
	end
	if (find("RESET", @w) = 0) notfound("RESET");
	patch(Coldjmp, w);
	if (find("H", @w) = 0) notfound("H");
	patch(w-ORG, Here+ORG);
	if (find("(FORTH)", @w) = 0) notfound("(FORTH)");
	patch(w-ORG, Last+ORG);
end

init() do
	Bp := 0;
	Nchars := 0;
	Sp := 0;
	Input := t.open("t3xforth.fth", T3X.OREAD);
	if (Input < 0)
		error("no source code found (t3xforth.fth)", 0);
end

cachewords() do
	DoColon := lookup("(COLON)", 0);
	DoVar := lookup("(VAR)", 0);
	DoConst := lookup("(CONST)", 0);
	DoBranch := lookup("(BRANCH)", 1);
	DoZBranch := lookup("(0BRANCH)", 1);
	DoLit := lookup("(LIT)", 1);
	DoDo := lookup("(DO)", 1);
	Exit := lookup("EXIT", 1);
	DoLoop := lookup("(LOOP)", 1);
	DoPlusLoop := lookup("(+LOOP)", 1);
	DoUpLoop := lookup("(/LOOP)", 1);
	DoString := 0;
end

do var src::14, dest::14, k, k2;
	k := t.getarg(1, src, 13);
	k2 := t.getarg(2, dest, 13);
	init();
	loadimage(k>0-> src: "t3xforth.bin");
	cachewords();
	loadsrc();
	listwords();
	saveimage(k2>0-> dest: "t3xforth.com");
end
